import os
import sys
from logging import DEBUG
from logging import basicConfig

from click import argument
from click import command
from click import echo
from click import get_current_context
from click import option
from click import style

from momake.exceptions import TaskFailed
from momake.finder import TaskFinder
from momake.runlog.commands import clear_database
from momake.task import Task


class Runner:
    def __init__(self):
        self.tasks: dict[str, Task] = {}

    def add(self, task: Task):
        assert task.name
        self.tasks[task.name] = task

    def run(self, taskname: str):
        try:
            self.tasks[taskname].run()
        except TaskFailed as er:
            ctx = get_current_context()
            echo(style(f"Task filed: {er.args[0]}", fg="red"))
            ctx.exit(1)

    def list_tasks(self):
        echo("Avalible tasks:")
        for task in self.tasks.values():
            echo(f"- {task.name}")


@command()
@argument("name", required=False, default=None)
@option("-l", "--list-tasks", default=False, is_flag=True, help="number of greetings")
@option("-c", "--clear", default=False, is_flag=True, help="clear database")
@option("-d", "--debug", default=False, is_flag=True, help="show debug logs")
def cmd(name: str, list_tasks: bool, clear: bool, debug: bool):
    sys.path.append(os.getcwd())
    if debug:
        basicConfig(level=DEBUG, format="%(message)s")

    if clear:
        clear_database()
        echo("Database deleted...")
        return

    runner = Runner()
    for taskcls in TaskFinder().find():
        task = taskcls()
        runner.add(task)

    if list_tasks:
        runner.list_tasks()
        return

    if name:
        runner.run(name)
        return 1

    ctx = get_current_context()
    echo(ctx.get_help())
    echo()
    runner.list_tasks()



from datetime import datetime
from datetime import timezone
from functools import cache
from pathlib import Path
from typing import Optional
from uuid import UUID


class Dependency:
    def has_changed(self, last_runtime: datetime, run_id: UUID) -> bool:
        return False

    def should_run(self, last_runtime: Optional[datetime], run_id: UUID) -> bool:
        if not last_runtime:
            return True
        return self.has_changed(last_runtime, run_id)


class AlwaysDependency(Dependency):
    def has_changed(self, last_runtime: datetime, run_id: UUID) -> bool:
        return True


class FileDependency(Dependency):
    def __init__(self, path: str, pattern: str):
        super().__init__()
        self.path = path
        self.pattern = pattern

    @cache
    def file_mtimes(self) -> list[datetime]:
        mtimes = []
        pathobj = Path(self.path)
        for path in pathobj.rglob(self.pattern):
            modified = datetime.fromtimestamp(path.lstat().st_mtime, tz=timezone.utc)
            mtimes.append(modified)
        return mtimes

    def has_changed(self, last_runtime: datetime, run_id: UUID) -> bool:
        if not self.file_mtimes():
            return True
        return min(self.file_mtimes()) > last_runtime



from datetime import datetime
from datetime import timezone
from typing import Optional
from uuid import UUID
from uuid import uuid4

from click import echo
from click import style

from momake.runlog.commands import set_task_log
from momake.runlog.queries import get_last_task_run


class Task:
    name = None
    dependecies = []

    @property
    def task_id(self):
        return self.__class__.__name__

    def should_run(self, last_runtime: Optional[datetime], run_id: UUID) -> bool:
        return self.run(run_id)

    def should_run_task(self, run_id: UUID):
        last_runtime = get_last_task_run(self.task_id)
        for dependency in self.dependecies:
            if dependency.should_run(last_runtime, run_id):
                return True
        return False

    def action(self):
        pass

    def run(self, run_id: Optional[UUID] = None):
        run_id = run_id or uuid4()
        echo(style(f"Checking {self.name}...", fg="bright_yellow"))
        runlog = {}
        runlog["check_runtime"] = datetime.now(timezone.utc)
        if self.should_run_task(run_id):
            runlog["runtime"] = datetime.now(timezone.utc)
            echo(style(f"Running {self.name}...", fg="green"))
            self.action()
            runlog["runtime_finish"] = datetime.now(timezone.utc)
            result = True
        else:
            result = False
        set_task_log(self.task_id, run_id, **runlog)
        return result



from importlib import import_module
from inspect import getmodule
from logging import getLogger
from pkgutil import walk_packages

from momake.task import Task

logger = getLogger(__name__)


def is_defined_in(element, module):
    return getmodule(element).__name__ == module.__name__


class TaskFinder:
    modules = ["momaketasks"]

    def find(self):
        elements = []
        for parent in self.modules:
            info = "Searching for all objects in %s"
            logger.info(info, parent)
            for package in self._get_all_packages(parent):
                elements += list(self._find_in_package(package))
        logger.info("")
        return elements

    def _get_all_packages(self, parent: str):
        try:
            parentpkg = import_module(parent)
        except Exception:
            logger.error("Can not import parent module: %s", parent)
            return

        prefix = f"{parentpkg.__name__}."
        modpath = getattr(parentpkg, "__path__", None)
        if not modpath:
            yield parentpkg
            return

        for module in walk_packages(modpath, prefix):
            try:
                yield import_module(module.name)
            except Exception:
                logger.warning("Can not import module: %s", module.name)
                continue

    def _find_in_package(self, package):
        intro = False
        for elementname in dir(package):
            element = getattr(package, elementname)

            if self.is_collectable(element) and is_defined_in(element, package):
                if not intro:
                    logger.debug("Module found: %s", package.__name__)
                    intro = True
                logger.debug(f"\tTask found: %s", elementname)
                yield element

    def is_collectable(self, element: object):
        try:
            return issubclass(element, Task)
        except TypeError:
            return False



from sqlalchemy import UUID
from sqlalchemy import Column
from sqlalchemy import DateTime
from sqlalchemy import String

from momake.runlog.db import SqlTable


class TaskTable(SqlTable):
    __tablename__ = "tasks"

    task_id = Column(String)
    run_id = Column(UUID)
    check_runtime = Column(DateTime(timezone=True))
    runtime = Column(DateTime(timezone=True))
    runtime_finish = Column(DateTime(timezone=True))



from datetime import datetime
from datetime import timezone
from typing import Optional
from uuid import UUID

from sqlalchemy import select
from sqlalchemy.orm import Session

from momake.runlog.conn import engine
from momake.runlog.tables import TaskTable


def get_last_task_run(task_id) -> Optional[datetime]:
    stmt = select(TaskTable.runtime).filter(
        TaskTable.task_id == task_id,
        TaskTable.runtime.isnot(None),
    ).order_by(TaskTable.runtime.desc())
    with Session(engine()) as db:
        result = db.execute(stmt).first()
        if result:
            return result.runtime.replace(tzinfo=timezone.utc)
    return None


def get_task(task_id: str, run_id: UUID) -> Optional[dict]:
    stmt = select(TaskTable).filter(
        TaskTable.task_id == task_id, TaskTable.run_id == run_id
    )
    with Session(engine()) as db:
        result = db.execute(stmt).first()
        if result:
            return result._asdict()
    return None



from functools import cache

from sqlalchemy import create_engine

import momake.runlog.tables
from momake.runlog.db import metadata

FILE_URL = "momake.db"
URL = f"sqlite:///{FILE_URL}"


@cache
def engine():
    engine = create_engine(URL)
    metadata.create_all(engine, checkfirst=True)
    return engine



from sqlalchemy import UUID
from sqlalchemy import Column
from sqlalchemy import DateTime
from sqlalchemy import String
from sqlalchemy.orm import declarative_base
from sqlalchemy.schema import MetaData

metadata = MetaData()


class Base:
    id = Column(
        UUID(as_uuid=True),
        primary_key=True,
        unique=True,
        nullable=False,
    )
    created_at = Column(DateTime(timezone=True))
    updated_at = Column(DateTime(timezone=True))

    def _asdict(self):
        data = dict(self.__dict__)
        del data["_sa_instance_state"]
        return data


SqlTable = declarative_base(cls=Base, metadata=metadata)



import os
from datetime import datetime
from datetime import timezone
from typing import Optional
from uuid import UUID
from uuid import uuid4

from sqlalchemy import insert
from sqlalchemy import update
from sqlalchemy.orm import Session

from momake.runlog.conn import FILE_URL
from momake.runlog.conn import engine
from momake.runlog.queries import get_task
from momake.runlog.tables import TaskTable


def set_task_log(
    task_id: str,
    run_id: UUID,
    check_runtime: Optional[datetime] = None,
    runtime: Optional[datetime] = None,
    runtime_finish: Optional[datetime] = None,
):
    task = get_task(task_id, run_id)
    if task:
        values = dict(
            check_runtime=check_runtime,
            runtime=runtime,
            runtime_finish=runtime_finish,
            updated_at=datetime.now(timezone.utc),
        )
        filtered = {k: v for k, v in values.items() if v is not None}
        stmt = (
            update(TaskTable)
            .where(
                TaskTable.task_id == task_id,
                TaskTable.run_id == run_id,
            )
            .values(**filtered)
        )
    else:
        stmt = insert(TaskTable).values(
            id=uuid4(),
            task_id=task_id,
            run_id=run_id,
            check_runtime=check_runtime,
            runtime=runtime,
            runtime_finish=runtime_finish,
            created_at=datetime.now(timezone.utc),
            updated_at=datetime.now(timezone.utc),
        )
    with Session(engine()) as db:
        db.execute(stmt)
        db.commit()


def clear_database():
    try:
        os.unlink(FILE_URL)
    except FileNotFoundError:
        pass
